#

import logging
import time
import codecs


## Monkey-patching parsimonious 0.8 to support Python 3.11

import sys
import inspect

# Check if we're running on Python 3.11 or later
if sys.version_info >= (3, 11):
    # Implement a getargspec function using getfullargspec for compatibility
    def getargspec(func):
        full_argspec = inspect.getfullargspec(func)
        args = full_argspec.args
        varargs = full_argspec.varargs
        varkw = full_argspec.varkw
        defaults = full_argspec.defaults
        return inspect.ArgSpec(args, varargs, varkw, defaults)

    # Monkey patch the inspect module
    inspect.getargspec = getargspec

## End of monkey-patching parsimonious


from web3 import Web3
from web3.eth import Eth
import eth_abi

from .wallet import Wallet
from .ApiCommunication import ApiCommunication
from .ApiHandling import ApiHandling, Endpoint

logger = logging.getLogger(__name__)


NONE = {}


def decode_data(abi_types, data):
    unique = False
    if isinstance(abi_types, str):
        unique = True
        abi_types = [abi_types]
    if data.startswith("0x"):
        data = data[2:]

    if len(data) == 0:
        return None

    try:
        data_buffer = bytes.fromhex(data)
    except ValueError:
        raise ValueError("Invalid data provided: not a hex string")

    if len(data_buffer) % 32 != 0:
        raise ValueError("Invalid data provided: data length is not a multiple of 32")

    res = eth_abi.decode(abi_types, data_buffer)
    if unique:
        return res[0]
    return res




class WalletLocked(Exception): pass


class AddressableObject:

    def __init__(self, address):
        self.address = address


class Account(AddressableObject): pass
class Transaction(AddressableObject): pass


class Pyc3l:

    def __init__(self, endpoint=None):
        self._additional_nonce = 0

        self._current_block = 0

        self._endpoint_last_usage = None

        if endpoint:
            logger.info(f"endpoint: {endpoint} (fixed)")
            self._endpoint = Endpoint(endpoint)
            self._endpoint_resolver = None
        else:
            self._endpoint = None
            self._endpoint_resolver = ApiHandling()

    @property
    def endpoints(self):
        if self._endpoint_resolver is not None:
            return self._endpoint_resolver.endpoints
        return [self._endpoint]

    @property
    def endpoint(self):
        if self._endpoint_resolver is not None:
            now = time.time()
            if self._endpoint and now - self._endpoint_last_usage > 2 * 60:
                self._endpoint = None
                logger.info("Re-selection of an endpoint triggered")
            self._endpoint_last_usage = now
            if self._endpoint is None:
                self._endpoint = self._endpoint_resolver.endpoint
                logger.info(f"endpoint: {self._endpoint} (elected)")
        return self._endpoint

    @property
    def ipfs_endpoint(self):
        if self._endpoint_resolver is not None:
            return self._endpoint_resolver.ipfs_endpoint
        return self._endpoint

    ## Blockchain information

    def getBlockNumber(self):
        return self.endpoint.api.post()

    def getTransactionBlock(self, transaction_hash):
        info = self.getTransactionInfo(transaction_hash)
        return info["transaction"]["blockNumber"]

    def getTransactionInfo(self, transaction_hash):
        data = {"hash": transaction_hash}

        r = self.endpoint.api.post(data=data)
        ## XXXvlab: seems to need to be parsed twice (confirmed upon
        ## reading the code of the comchain API).
        if isinstance(r, str):
            import json

            r = json.loads(r)
        return r

    def getTrInfos(self, address):
        return self.endpoint.api.post(data={"txdata": address})

    def getAccountEthBalance(self, address):
        return self.endpoint.api.post(data={"balance": address})['balance']

    def hasChangedBlock(self, do_reset=False):
        new_current_block = self.getBlockNumber()
        res = new_current_block != self._current_block
        if do_reset:
            self._current_block = new_current_block
        return res

    def registerCurrentBlock(self):
        self.hasChangedBlock(do_reset=True)

    def read(self, fn, args, abi_return_type="int256"):
        data = {
            "ethCall": {
                "to": fn[0],  ## contract
                "data": fn[1] + "".join(
                    (v[2:] if v.startswith("0x") else v).zfill(64)
                    for v in args
                )
            }
        }
        try:
            result = self.endpoint.api.post(data=data)
        except Exception as e:
            logger.error(
                "Unexpected failure of ethCall " +
                f"contract: 0x{fn[0]}, fn: 0x{fn[1]}, args: {args!r}"
            )
            raise e
        if abi_return_type is None:
            return result
        return decode_data(abi_return_type, result)

    def get_element_in_list(self, map_fn, amount_fn, caller_address,
                            idx, dct, idx_min):
        if idx < idx_min:
            return dct
        data = self.read(map_fn, [caller_address, hex(idx)], None)
        amount = self.read(amount_fn, [caller_address, data]) / 100.0

        address = "0x" + data[-40:]
        dct[address] = amount
        return self.get_element_in_list(
            map_fn, amount_fn,
            caller_address, idx - 1, dct, idx_min
        )

    ## Blockchain transaction

    def send_transaction(
            self,
            fn,
            data,
            account,
            ciphered_message_from="",
            ciphered_message_to="",
    ):
        tr_infos = self.getTrInfos(account.address)
        gas_price = int(tr_infos["gasprice"], 0)
        gas_price_gwei = Web3.fromWei(gas_price, "gwei")
        nonce = int(tr_infos["nonce"], 0)
        logger.info(f"Gas price: {gas_price!r} wei ({gas_price_gwei} gwei), Nonce: {nonce!r}")
        transaction = {
            "to": fn[0],
            "value": 0,
            # "gas": 2500000,
            "gas": 5000000,
            "gasPrice": gas_price,
            "nonce": self.update_nonce(nonce),
            "data": fn[1] + data,
            "from": account.address,
        }

        signed = Eth.account.signTransaction(transaction, account.privateKey)
        str_version = (
            "0x" + str(codecs.getencoder("hex_codec")(signed.rawTransaction)[0])[2:-1]
        )
        raw_tx = {"rawtx": str_version}

        if ciphered_message_from != "":
            raw_tx["memo_from"] = ciphered_message_from

        if ciphered_message_to != "":
            raw_tx["memo_to"] = ciphered_message_to

        return self.endpoint.api.post(data=raw_tx)

    def update_nonce(self, nonce):
        if not self.hasChangedBlock(do_reset=True):
            self._additional_nonce = self._additional_nonce + 1
            return nonce + self._additional_nonce
        else:
            self._additional_nonce = 0
            return nonce

    ## Sub-objects

    def Currency(pyc3l_instance, name):
        class Pyc3lCurrency(ApiCommunication):

            def Account(pyc3l_currency, address):

                class Pyc3lCurrencyAccount(Account):

                    def __getattr__(self, label):
                        label = label[0].upper() + label[1:]
                        if label.startswith("getAccount"):
                            method = getattr(pyc3l_currency, label, NONE)
                            if method is not NONE:
                                return lambda: method(self.address)
                        method = getattr(pyc3l_currency, f"getAccount{label}", NONE)
                        if method is NONE:
                            method = getattr(pyc3l_instance, f"getAccount{label}", NONE)
                        if method is not NONE:
                            return method(self.address)
                        if label.endswith("s"):
                            method = getattr(pyc3l_currency, f"get{label[:-1]}List", NONE)
                            if method is not NONE:
                                return method(self.address)
                        raise AttributeError(label)

                return Pyc3lCurrencyAccount(address)

            @property
            def symbol(self):
                return self.metadata["server"]["currencies"]["CUR"]

        return Pyc3lCurrency(name, pyc3l_instance)

    @property
    def Wallet(pyc3l_instance):

        class Pyc3lWallet(Wallet):

            @property
            def currency(self):
                return pyc3l_instance.Currency(self._wallet["server"]["name"])

            @property
            def account(self):
                return self.currency.Account(self.address)

            def __getattr__(self, label):
                method = getattr(self.account, label, NONE)
                if method is not NONE:
                    return method
                if label.startswith("transfer") or label in ["lockUnlockAccount", "pledge",
                                                             "delegate", "transferOnBehalfOf",
                                                             "enable", "disable"]:
                    if not self._account:
                        raise WalletLocked("Wallet is required to be unlocked")
                    method = getattr(self.currency, label, NONE)
                    if method is not NONE:
                        return lambda *args, **kwargs: method(self._account, *args, **kwargs)
                raise AttributeError(label)

        return Pyc3lWallet

    def Transaction(pyc3l_instance, address):

        class Pyc3lTransaction(Transaction):

            @property
            def data(self):
                return pyc3l_instance.getTransactionInfo(self.address)

            @property
            def block(self):
                return pyc3l_instance.getTransactionBlock(self.address)

            @property
            def pending(self):
                return self.block is None

        return Pyc3lTransaction(address)




import requests

import random
import datetime
import os.path
import logging
import tempfile
import os
import shutil
import traceback
import time
import itertools

from contextlib import closing

logger = logging.getLogger(__name__)


class HTTPError(Exception): pass

class APIError(Exception): pass


def urlencode_prepare_dict(dct):
    """Flattens nested dict to support urlencode

    Non nested dicts are unchanged:
    
    >>> urlencode_prepare_dict({"a": 1})
    {'a': 1}

    But nested dicts will be flatten to be encoded in URL
    or ``application/x-www-form-urlencode`` POST bodies:

    >>> urlencode_prepare_dict({"a": {"b": 1}})
    {'a[b]': 1}

    >>> urlencode_prepare_dict({"a": {"b": 1}, "c": 2})
    {'a[b]': 1, 'c': 2}

    """
    def flatten_dict(d, parent_key=""):
        nd = {}
        for k, v in d.items():
            new_key = f"{parent_key}[{k}]" if parent_key else k
            if isinstance(v, dict):
                nd.update(flatten_dict(v, new_key))
            else:
                nd[new_key] = v
        return nd

    return flatten_dict(dct)


class BaseEndpoint(object):
    """Simple request shortcut

    This object keeps the base url of an endpoint and allows
    to call `requests` by specifying only the `path` part:

    Let's setup the mock to see how it calls `requests`:

        >>> import minimock
        >>> minimock.mock('requests.get')
        >>> minimock.mock('requests.post')

    Instantiate by specifying the base url:

        >>> e = BaseEndpoint('http://example.com')
        >>> e.get('/path', 1, 2, foo="bar")
        Called requests.get('http://example.com/path', 1, 2, foo='bar')
        >>> e.post('/path', 1, 2, foo="bar")
        Called requests.post('http://example.com/path', 1, 2, foo='bar')

    We can see that positional arguments and keyword arguments are
    correctly sent to `requests.*` methods.

        >>> minimock.restore()

    """

    def __init__(self, url):
        self._url = url

    def __getattr__(self, label):
        if label not in ["get", "post"]:
            raise AttributeError()

        def r(*args, **kwargs):
            if len(args):
                args = list(args)
                args[0] = f"{self._url}{args[0]}"
            else:
                args = [f"{self._url}"]
            logger.debug("Request %s %s %r" % (
                label.upper(),
                args[0],
                kwargs
            ))
            if "data" in kwargs:
                kwargs["data"] = urlencode_prepare_dict(kwargs["data"])
            res = getattr(requests, label)(*args, **kwargs)
            logger.debug("  Response [%s]: %d bytes" % (
                res.status_code,
                len(res.text),
            ))
            if res.status_code != 200:
                raise HTTPError("%s %s ERROR (%s)" % (
                    label.upper(),
                    args[0],
                    res.status_code
                    ))
            try:
                data = res.json()
            except Exception:
                raise Exception("%s %s ERROR (Not JSON data)" % (
                    label.upper(),
                    self._url,
                ))

            if not isinstance(data, dict):
                return data

            if data.get("error", False):
                data_msg = f" (data: {data.get('data')})" if data['data'] else ""
                if data['msg']:
                    raise APIError(f"API Call failed with message: {data['msg']}{data_msg}")
                raise APIError("API Call failed without message{data_msg}")
            if "data" in data:
                return data["data"]
            return data

        return r


class Endpoint(BaseEndpoint):

    URLS = {
        "api": "/api.php",
        "config": "/ipns/QmaAAZor2uLKnrzGCwyXTSwogJqmPjJgvpYgpmtz5XcSmR/configs/",
        "endpoint_list": "/ipns/QmcRWARTpuEf9E87cdA4FfjBkv7rKTJyfvsLFTzXsGATbL",
        "keys": "/keys.php",
        "transactions": "/trnslist.php",
    }

    def __init__(self, url):
        self._url = url

    def __getattr__(self, label):
        if label in ["get", "post"]:
            return super(Endpoint, self).__getattr__(label)
        if label in self.URLS.keys():
            return BaseEndpoint(f"{self._url}{self.URLS[label]}")
        raise AttributeError()

    def __repr__(self):
        return f"<{self.__class__.__name__} {self._url!r}>"

    def __str__(self):
        return self._url

    def __hash__(self):
        return hash(self._url)

    def __eq__(self, value):
        return isinstance(value, Endpoint) and self._url == value._url


def random_picker(elts):
    """Iterator giving endless random pick in given set of elts

    Don't forget to set the seed if you expect to give you a different
    set after recreating the iterator.

    """
    while True:
        yield random.choice(elts)


def first(elts, predicate):
    return next(filter(predicate, elts))


def first_pick(elts, predicate, max_retries=20):
    return first(
        itertools.islice(elts, 0, max_retries),
        predicate,
    )


class ApiHandling(object):

    DEFAULT_ENDPOINTS = [
        "https://node-cc-001.cchosting.org",
        "https://node-001.cchosting.org",
        "https://node-002.cchosting.org",
        "https://node-003.cchosting.org",
        "https://node-004.cchosting.org",
    ]
    UPDATE_INTERVAL = 60 * 15  ## in sec

    def __init__(self, endpoint_file=None, max_retries=20):
        self._store = (
            SimpleFileStore(endpoint_file)
            if endpoint_file
            else StateFileStore("pyc3l", "endpoints.txt")
        )
        self._mtime = 0  ## cache and lazy loading
        self._max_retries = max_retries
        self._endpoints = None

    @property
    def endpoints(self):
        if not self._endpoints:
            self._endpoints, self._mtime = self._load()
        if self._endpoints is None:
            self._endpoints = set(Endpoint(e) for e in self.DEFAULT_ENDPOINTS)
        if time.time() - self._mtime < self.UPDATE_INTERVAL:
            return self._endpoints

        logger.info("endpoint list is obsolete, requesting update of list")
        if not self._update(force=True):
            logger.warn("endpoint list update failed, keeping old list")

        return self._endpoints

    def _update(self, force=False):
        logger.info("updating endpoint list")
        ## Avoid to use self.endpoints for recursivity reasons
        endpoints = (
            self._endpoints
            or self._load()[0]
            or set(Endpoint(e) for e in self.DEFAULT_ENDPOINTS)
        )
        for endpoint in random.sample(list(endpoints), k=len(endpoints)):
            logger.debug("  Try to get endpoint list from %r", endpoint)
            r = self._safe_req(endpoint.endpoint_list.get)
            if r is not False:
                break

        if r is False:
            logger.error("Failed to get new endpoint list.")
            return False

        logger.info("  Got endpoint list from %r", endpoint)
        new_endpoints = set(Endpoint(e[:-1] if e[-1] == "/" else e) for e in r)
        modified = endpoints != new_endpoints
        if not force and not modified:
            logger.info("saved endpoint list is already up-to-date.")
            return

        if modified:
            removed = endpoints - new_endpoints
            added = new_endpoints - endpoints
            msg = "\n".join(
                sorted(["  - %s" % e for e in removed] + ["  + %s" % e for e in added])
            )
            logger.info("  Updating endpoint list:\n%s", msg)
        else:
            logger.info("  Update endpoint last modification time.")

        self._save(new_endpoints)
        return True

    def _first_pick_endpoints(self, predicate, max_retries):
        random.seed(time.time())
        endpoint = first_pick(random_picker(list(self.endpoints)), predicate, max_retries)
        if endpoint:
            return endpoint
        raise Exception(
            f"No endpoint found able to fullfill predicate after {max_retries} retries."
        )

    @property
    def ipfs_endpoint(self):
        return self._first_pick_endpoints(
            lambda e: self._safe_req(e.config.get, "ping.json") is not False, self._max_retries
        )

    @property
    def endpoint(self):
        return self._first_pick_endpoints(
            lambda e: self._safe_req(e.api.get), self._max_retries
        )

    def _safe_req(self, method, path=""):
        try:
            r = method(f"{path}?_={datetime.datetime.now()}")
        except Exception as e:
            logger.warn("request raised exception: %s", e)
            return False
        return r

    def _save(self, endpoints):
        self._store.save("\n".join(["%s" % e for e in endpoints]))
        self._endpoints = endpoints
        self._mtime = time.time()

    def _load(self):
        saved_data, last_mtime = self._store.load(with_mtime=True)
        if saved_data is None:
            return set(), last_mtime
        endpoints = set(
            Endpoint(e[:-1] if e[-1] == "/" else e)
            for e in saved_data.strip().split("\n")
        )
        logger.info(
            "read local list of endpoint: %d endpoints, last mtime: %s",
            len(endpoints),
            datetime.datetime.utcfromtimestamp(last_mtime).isoformat(),
        )
        return (
            endpoints,
            last_mtime,
        )


class SimpleFileStore:
    """Manage saving and loading text data in file

    The class constructor takes a filename as input, which represents
    the name of the file where data will be saved.

        >>> import tempfile
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     store = SimpleFileStore(f"{tmpdir}/data.txt")
        ...     store.save("Hello, World!")
        ...     store.load()
        'Hello, World!'

    Take care of creating directory if not existent.

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     store = SimpleFileStore(f"{tmpdir}/a/b/c/data.txt")
        ...     store.save("Hello, World!")
        ...     store.load()
        'Hello, World!'

    Creation of the directory happens on save only.

        >>> import os
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     store = SimpleFileStore(f"{tmpdir}/data.txt")
        ...     os.path.isfile(f"{tmpdir}/data.txt")
        False

    Load will return ``None`` if file is not existent (nothing was never saved).

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     store = SimpleFileStore(f"{tmpdir}/a/b/c/data.txt")
        ...     store.load() is None
        True

    Ensure atomicity by using a temporary file and use move (atomic)
    to replace the target file. This also ensure that only fully
    written file will be saved.

    - This class is best used for small to medium-sized files since it
      reads and writes the entire file into memory at once.

    - The class assumes that the file will contain text data. If
      binary data needs to be saved, the file should be opened in
      binary mode.

    """

    def __init__(self, filename):
        self._filename = filename

    def save(self, content):
        """Save endpoints in file"""
        f, tmp = tempfile.mkstemp()
        with closing(os.fdopen(f, "w")) as file:
            file.write(content)
        if not os.path.isfile(self._filename):
            logger.info("Create local endpoint file named %r.", self._filename)
            os.makedirs(os.path.dirname(self._filename), exist_ok=True)
        shutil.move(tmp, self._filename)  ## atomic

    def load(self, with_mtime=False):
        if not os.path.isfile(self._filename):
            return (None, 0) if with_mtime else None
        with open(self._filename, "r") as file:
            data = file.read()
        return (data, os.path.getmtime(self._filename)) if with_mtime else data


class StateFileStore(SimpleFileStore):
    """Manage saving and loading data in file in state directory

     It will use XDG_STATE_HOME environment variable if available

         >>> import tempfile, os
         >>> with tempfile.TemporaryDirectory() as tmpdir:
         ...     os.environ["XDG_STATE_HOME"] = tmpdir
         ...     store = StateFileStore("bar", "foo")
         ...     store._filename == f"{tmpdir}/bar/foo"
         True

     Otherwise will fallback on ``~/.local/state/{group}/{name}``

         >>> with tempfile.TemporaryDirectory() as tmpdir:
         ...     del os.environ["XDG_STATE_HOME"]
         ...     os.environ["HOME"] = tmpdir
         ...     store = StateFileStore("bar", "foo")
         ...     store._filename == f"{tmpdir}/.local/state/bar/foo"
         True

    It'll get all the features of SimpleFileStore:

         >>> import tempfile
         >>> with tempfile.TemporaryDirectory() as tmpdir:
         ...     os.environ["XDG_STATE_HOME"] = tmpdir
         ...     store = StateFileStore("bar", "data.txt")
         ...     assert not os.path.isfile(store._filename)  ## doesn't exist yet
         ...     store.save("Hello, World!")
         ...     assert os.path.isfile(store._filename)      ## created
         ...     store.load()
         'Hello, World!'


    """

    def __init__(self, group, name):
        xdg_state_home = os.environ.get("XDG_STATE_HOME") or os.path.join(
            os.path.expanduser("~"), ".local", "state"
        )

        super(StateFileStore, self).__init__(os.path.join(xdg_state_home, group, name))


def format_last_exception(prefix="  | "):
    """Format the last exception for display it in tests.

    This allows to raise custom exception, without loosing the context of what
    caused the problem in the first place:

    >>> def f():
    ...     raise Exception("Something terrible happened")
    >>> try:  ## doctest: +ELLIPSIS
    ...     f()
    ... except Exception:
    ...     formated_exception = format_last_exception()
    ...     raise ValueError('Oups, an error occured:\\n%s' % formated_exception)
    Traceback (most recent call last):
    ...
    ValueError: Oups, an error occured:
      | Traceback (most recent call last):
    ...
      | Exception: Something terrible happened

    """

    return "\n".join(
        str(prefix + line) for line in traceback.format_exc().strip().split("\n")
    )



from Crypto import Random
from Crypto.Cipher import AES
from Crypto.Hash import SHA512
from ecdsa import SigningKey, SECP256k1, ECDH, VerifyingKey

import struct
import hmac


## Code adapted from https://github.com/LimelabsTech/eth-ecies
## Match (tested against) the JS code of Biletujo


BS = 16


def pad(data):
    padding = BS - len(data) % BS
    return data + padding * struct.pack("B", padding)


def unpad(data):
    return data[0 : -data[-1]]


def AES256CbcDecrypt(hex_data, key=b"0" * 32, iv=b"0" * 16):
    data = bytearray.fromhex(hex_data)
    aes = AES.new(key, AES.MODE_CBC, iv)
    return unpad(aes.decrypt(data))


def AES256CbcEncrypt(bin_data, key=b"0" * 32, iv=b"0" * 16):
    aes = AES.new(key, AES.MODE_CBC, iv)
    return aes.encrypt(pad(bin_data))


def EncryptMessage(publicKey_hex, plainText_string):
    if publicKey_hex[:2] == "0x":
        publicKey_hex = publicKey_hex[2:]
    publicKey_bin = bytearray.fromhex(publicKey_hex)

    # Generate the temporary key
    ecdh = ECDH(curve=SECP256k1)
    ecdh.generate_private_key()

    ephemPubKeyEncoded = bytearray.fromhex(ecdh.get_public_key().to_string().hex())

    # Load the public key
    publicKey = VerifyingKey.from_string(publicKey_bin, curve=SECP256k1)

    # ECDH => get the shared secret
    ecdh.load_received_public_key(publicKey)

    px = ecdh.generate_sharedsecret_bytes()

    # compute the encription and MAC keys
    hash_px = SHA512.new(data=px).digest()
    encryptionKey = hash_px[:32]
    macKey = hash_px[32:]

    # cipher the plain text
    iv = Random.get_random_bytes(16)
    plaintext = plainText_string.encode(encoding="utf_8")
    ciphertext = AES256CbcEncrypt(plaintext, encryptionKey, iv)

    # compute the MAC
    dataToMac = iv + bytearray([4]) + ephemPubKeyEncoded + ciphertext
    mac = hmac.new(macKey, dataToMac, "sha256").digest()

    # build the output
    serializedCiphertext = iv + bytearray([4]) + ephemPubKeyEncoded + mac + ciphertext
    return serializedCiphertext.hex()


def DecryptMessage(privateKey_hex, encrypted_hex):
    if privateKey_hex[:2] == "0x":
        privateKey_hex = privateKey_hex[2:]
    if encrypted_hex[:2] == "0x":
        encrypted_hex = encrypted_hex[2:]

    # get the components
    encrypted = bytearray.fromhex(encrypted_hex)
    iv = encrypted[:16]
    ephemPubKeyEncoded = encrypted[17:81]
    mac = encrypted[81:113]
    ciphertext = encrypted[113:]

    # recover the temporary public key
    ephemPubKey = VerifyingKey.from_string(ephemPubKeyEncoded, curve=SECP256k1)

    # load the private key
    priv_key = SigningKey.from_secret_exponent(int(privateKey_hex, 16), curve=SECP256k1)
    ecdh = ECDH(curve=SECP256k1, private_key=priv_key)

    # ECDH => get the shared secret
    ecdh.load_received_public_key(ephemPubKey)
    px = ecdh.generate_sharedsecret_bytes()

    # compute the encription and MAC keys
    hash_px = SHA512.new(data=px).digest()
    encryptionKey = hash_px[:32]
    macKey = hash_px[32:]

    # check the MAC
    dataToMac = iv + bytearray([4]) + ephemPubKeyEncoded + ciphertext
    computed_mac = hmac.new(macKey, dataToMac, "sha256").digest()
    if computed_mac != mac:
        raise ValueError("MAC missmatch")

    # decipher the text
    plaintext = AES256CbcDecrypt(ciphertext.hex(), encryptionKey, iv)
    return plaintext.decode("utf-8")



import logging
import re
from typing import NewType


from .CryptoAsim import EncryptMessage, DecryptMessage


logger = logging.getLogger(__name__)


Address = NewType('Address', str)
Uint256 = NewType('Uint256', int)
Amount = NewType('Amount', Uint256)
Bool = NewType('Bool', Uint256)
Bytes = NewType('Bytes', bytes)



def encodeNumber(number):
    if number < 0:
        return hex(16**64 + number)[2:].zfill(64)
    else:
        return str(hex(number))[2:].zfill(64)


def encodeAddressForTransaction(address):
    full_address = address
    if full_address.startswith("0x"):
        full_address = full_address[2:]
    if len(full_address) != 40:

        raise Exception("Missformed wallet address: " + address)
    return full_address.zfill(64)


class ComChainABI:

    def accountType(account: Address) -> Uint256: "ba99af70"
    def accountIsActive(account: Address) -> Bool: "61242bdd"
    def accountIsOwner(account: Address) -> Bool: "2f54bf6e"
    def accountCmLimitMax(account: Address) -> Amount: "cc885a65"
    def accountCmLimitMin(account: Address) -> Amount: "ae7143d6"
    def accountGlobalBalance(account: Address) -> Amount: "70a08231"
    def accountNantBalance(account: Address) -> Amount: "ae261aba"
    def accountCmBalance(account: Address) -> Amount: "bbc72a17"

    def setAccountParam(
            address: Address,
            status: Uint256,
            accountType: Uint256,
            limitP: Amount,
            limitM: Amount): "848b2592"
    def pledge(address: Address, amount: Amount): "6c343eef"
    def delegate(address: Address, amount: Amount): "75741c79"

    def nantTransfer(dest: Address, amount: int): "a5f7c148-1"
    def cmTransfer(dest: Address, amount: int): "60ca9c4c-1"
    def transferNantOnBehalf(src: Address, to: Address, amount: int): "1b6b1ee5-1"

    def accountAllowances(account: Address) -> list[Amount]: "aa7adb3d-1,b545b11f-1,dd62ed3e-1"
    def accountRequests(account: Address) -> list[Amount]: "debb9d28-1,726d0a28-1,3537d3fa-1"
    def accountMyRequests(account: Address) -> list[Amount]: "418d0fd4-1,0becf93f-1,09a15e43-1"
    def accountDelegations(account: Address) -> list[Amount]: "58fb5218-1,ca40edf1-1,046d3307-1"
    def accountMyDelegations(account: Address) -> list[Amount]: "7737784d-1,49bce08d-1,f24111d2-1"
    def accountAcceptedRequests(account: Address) -> list[Amount]: "8d768f84-1,59a1921a-1,958cde37-1"
    def accountRejectedRequests(account: Address) -> list[Amount]: "20cde8fa-1,9aa9366e-1,eac9dd4d-1"


class Contract:

    def __init__(self, pyc3l, abi, contracts):
        self._pyc3l = pyc3l
        self._abi = abi
        self._contracts = contracts


    CONVERSIONS = {
        Uint256: lambda x: int(x),
        Amount: lambda x: x / 100.0,
        Bool: lambda x: bool(x),
    }

    @property
    def abi_read_functions(self):
        if not hasattr(self, "_abi_read_functions"):
            ## Read function have return value type annotations
            self._abi_read_functions = dict(filter(
                lambda x: x[1] is not None, [
                    (key, fn.__annotations__.get("return"))
                    for key, fn in self._abi.__dict__.items()
                    if callable(fn) and hasattr(fn, "__annotations__")
                ]))
        return self._abi_read_functions


    def _get_contract_fn_hexs(self, fn_name):
        fn_def = getattr(self._abi, fn_name, None)
        if fn_def is None:
            raise AttributeError(f"Function {fn_name!r} not found in ABI")
        fn_hexs = fn_def.__doc__.split(",")
        parsed_fn_hexs = []
        for fn_hex in fn_hexs:
            hex_parts = fn_hex.split("-", 1)
            if not re.match("^[0-9a-f]{8}$", hex_parts[0]):
                raise ValueError(
                    f"Invalid hex data provided ({hex_parts[0]!r}) for {fn_name!r} function"
                )

            contract = self._contracts[0 if len(hex_parts) == 1 else int(hex_parts[1])]
            fn_hex = f"0x{hex_parts[0]}"
            parsed_fn_hexs.append((contract, fn_hex))
        return parsed_fn_hexs

    def __getattr__(self, label):
        if not label.startswith("get"):
            raise AttributeError(label)

        key = label[3:]
        key = key[0].lower() + key[1:]

        return_type = self.abi_read_functions.get(key)
        if not return_type:
            raise AttributeError(
                f"Comchain read function {key!r} not found in ABI"
            )
        parsed_fn_hexs = self._get_contract_fn_hexs(key)

        if getattr(return_type, "__origin__", None) is list:
            try:
                count_fn, map_fn, amount_fn  = parsed_fn_hexs
            except ValueError:
                raise ValueError(
                    f"Invalid list function {key} docstring: {doc!r}, "
                    "please provide count, map, amount hex functions separated by commas"
                )
            def get_list_function(address, idx_min=0, idx_max=0):
                count = self._pyc3l.read(count_fn, [address])
                return self._pyc3l.get_element_in_list(
                    map_fn,
                    amount_fn,
                    address,
                    min(count - 1, idx_max),
                    {},
                    idx_min
                )

            return get_list_function


        def _method(address):
            value = self._pyc3l.read(parsed_fn_hexs[0], [address])
            return self.CONVERSIONS[return_type](value)
        return _method


class ApiCommunication:

    def __init__(self, currency_name, pyc3l, abi=ComChainABI):
        self._currency_name = currency_name

        self._current_block = 0
        self._metadata = None
        self._pyc3l = pyc3l

        self._comchain = None
        self._abi = abi

    @property
    def comchain(self):
        if self._comchain is None:
            self._comchain = Contract(
                self._pyc3l,
                self._abi,
                self.contracts
            )
        return self._comchain

    @property
    def endpoint(self):
        return self._pyc3l.endpoint

    @property
    def ipfs_endpoint(self):
        return self._pyc3l.ipfs_endpoint

    @property
    def metadata(self):
        if self._metadata is None:
            self._metadata = self._pyc3l.ipfs_endpoint.config.get(f"{self._currency_name}.json")
        return self._metadata

    @property
    def contracts(self):
        server = self.metadata["server"]
        return (
            server["contract_1"],
            server["contract_2"],
        )

    def checkAdmin(self, address):
        if not self.getAccountIsValidAdmin(address):
            raise Exception(
                f"The provided account {address} is not an "
                f"active admin on f{self._currency_name}"
                f" ({self.contracts[0]})"
            )

        if not self.getAccountHasEnoughGas(address):
            raise Exception(
                f"The provided account {address} has not enough gas."
            )



    ############################### messages with transaction handling
    def getMessageKeys(self, address, with_private):
        params = {"addr": address}
        if with_private:
            params["private"] = 1
        return self.endpoint.keys.get(params=params)

    def encryptTransactionMessage(self, plain_text, **kwargs):
        # if public_message_key is present use it if not get the key from the address
        if "public_message_key" in kwargs:
            public_message_key = kwargs["public_message_key"]
        elif "address" in kwargs:
            response = self.getMessageKeys(kwargs["address"], False)
            if "public_message_key" not in response:
                logger.warn("No message key for account %s", kwargs["address"])
                return "", ""
            public_message_key = response["public_message_key"]
        else:
            raise ValueError("public_message_key or address agrgument must be present")

        ciphered = EncryptMessage(public_message_key, plain_text)
        return ciphered, public_message_key

    def decryptTransactionMessage(self, ciphered, **kwargs):
        # if private_message_key is present use it if not get the key from the address and private_key
        if "private_message_key" in kwargs:
            private_message_key = kwargs["private_message_key"]
        elif "address" in kwargs and "private_key" in kwargs:
            response, r = self.getMessageKeys(kwargs["address"], True)
            if "private_message_key" not in response:
                logger.warn("No message key for account %s", kwargs["address"])
                return "", ""
            private_message_key = DecryptMessage(
                kwargs["private_key"], response["private_message_key"]
            )

        else:
            raise ValueError(
                "private_message_key or ( address and private_key ) agrgument must be present"
            )

        plain_text = DecryptMessage(private_message_key, ciphered)
        return plain_text, private_message_key

    ############################### High level Functions

    def getAccountIsValidAdmin(self, address):
        return (
            self.getAccountType(address) == 2
            and self.getAccountIsActive(address) == True
        )

    def getAccountHasEnoughGas(self, address, min_gas=5000000):
        return int(self._pyc3l.getTrInfos(address)["balance"]) > min_gas

    def getAccountTransactions(self, address, count=10, offset=0):
        transactions = self.endpoint.transactions.get(params={
            "addr": f"0x{address}",
            "count": count,
            "offset": offset,
        })
        ## XXXvlab: seems to need to be parsed twice (confirmed upon
        ## reading the code of the comchain API).
        import json
        return [json.loads(r) for r in transactions]

    ############################### High level Transactions
    def transferNant(self, account, dest_address, amount, **kwargs):
        # message_from="", message_to=""):
        """Transfer Nantissed current Currency (server) from the sender to the destination wallet

        Parameters:
        account (eth_account import::Account): An account with enough balance on the current server. Will sign the transaction
        dest_address (string): The public address of the wallet to be credited (0x12345... format)
        amount (double): amount (in the current Currency) to be transfered from the sender wallet to the destination wallet

        """

        # prepare messages
        if "message_from" in kwargs and kwargs["message_from"] != "":

            ciphered_message_from, public_key = self.encryptTransactionMessage(
                kwargs["message_from"], address=account.address
            )
        else:
            ciphered_message_from = ""

        if "message_to" in kwargs and kwargs["message_to"] != "":
            ciphered_message_to, public_key = self.encryptTransactionMessage(
                kwargs["message_to"], address=dest_address
            )
        else:
            ciphered_message_to = ""

        # Get sender wallet infos
        if not self.getAccountIsActive(account.address):
            raise Exception(
                f"The sender wallet {account.address} is locked "
                f"on {self._currency_name} ({self.contracts[0]}) "
                "and therefore cannot initiate a transfer."
            )

        balance = self.getAccountNantBalance(account.address)
        if balance < amount:
            raise Exception(
                f"The sender wallet {account.address} has an "
                f"insufficient Nant balance on {self._currency_name} "
                f"({self.contracts[0]}) to complete this transfer."
            )

        # Get destination wallet infos
        if not self.getAccountIsActive(dest_address):
            raise Exception(
                f"The destination wallet {dest_address} is locked on "
                f"{self._currency_name}  ({self.contracts[0]}) and "
                "therefore cannot receive a transfer."
            )

        # Prepare data
        amount_cent = int(100 * amount)
        data = encodeAddressForTransaction(dest_address)
        data += encodeNumber(amount_cent)

        # send transaction
        logger.info(
            "Transferring %s nantissed %s from wallet %s to target wallet %s on server %s",
            amount,
            self._currency_name,
            account.address,
            dest_address,
            "%s(%s, %s)" % (self._currency_name, self.contracts[0], self.contracts[1]),
        )
        return self._pyc3l.send_transaction(
            self.comchain._get_contract_fn_hexs("nantTransfer")[0],
            data,
            account,
            ciphered_message_from,
            ciphered_message_to,
        )

    def transferCM(self, account, dest_address, amount, **kwargs):
        # message_from="", message_to=""):
        """Transfer Mutual Credit current Currency (server) from the sender to the destination wallet

        Parameters:
        account (eth_account import::Account): An account with enough balance on the current server. Will sign the transaction
        dest_address (string): The public address of the wallet to be credited (0x12345... format)
        amount (double): amount (in the current Currency) to be transfered from the sender wallet to the destination wallet

        """
        # prepare messages
        if "message_from" in kwargs and kwargs["message_from"] != "":
            ciphered_message_from, public_key = self.encryptTransactionMessage(
                kwargs["message_from"], address=account.address
            )
        else:
            ciphered_message_from = ""

        if "message_to" in kwargs and kwargs["message_to"] != "":
            ciphered_message_to, public_key = self.encryptTransactionMessage(
                kwargs["message_to"], address=dest_address
            )
        else:
            ciphered_message_to = ""

        # Get sender wallet infos
        if not self.getAccountIsActive(account.address):
            raise Exception(
                "The sender wallet "
                + account.address
                + " is locked on server "
                + self._currency_name
                + " ("
                + self.contracts[0]
                + ") and therefore cannot initiate a transfer."
            )

        balance = self.getAccountCMBalance(account.address)
        if balance < amount:
            raise Exception(
                "The sender wallet "
                + account.address
                + " has an insuficient CM balance on server "
                + self._currency_name
                + " ("
                + self.contracts[0]
                + ") to complete this transfer."
            )

        # Get destination wallet infos
        if not self.getAccountIsActive(dest_address):
            raise Exception(
                "The destination wallet "
                + dest_address
                + " is locked on server "
                + self._currency_name
                + " ("
                + self.contracts[0]
                + ") and therefore cannot recieve a transfer."
            )

        # Prepare data
        amount_cent = int(100 * amount)
        data = encodeAddressForTransaction(dest_address)
        data += encodeNumber(amount_cent)

        # send transaction
        logger.info(
            "Transferring %s mutual credit %s from wallet %s to target wallet %s on server %s",
            amount,
            self._currency_name,
            account.address,
            dest_address,
            "%s(%s, %s)" % (self._currency_name, self.contracts[0], self.contracts[1]),
        )

        return self._pyc3l.send_transaction(
            self.comchain._get_contract_fn_hexs("cmTransfer")[0],
            data,
            account,
            ciphered_message_from,
            ciphered_message_to,
        )

    ############################### High level Admin restricted Transactions
    def enable(self, account, address):
        return self._lockUnlockAccount(account, address, lock=False)

    def disable(self, account, address):
        return self._lockUnlockAccount(account, address, lock=True)

    def _lockUnlockAccount(self, account, address, lock=True):
        """Lock or unlock an Wallet on the current Currency (server)

        Parameters:
        account (eth_account import::Account): An account with admin permission on the current server. Will sign the transaction
        address (string): The public address of the wallet to be locked/unlocked (0x12345... format)
        lock (bool): if True, lock the wallet, if False unlock it

        """
        # Check the admin
        self.checkAdmin(account.address)

        # Get wallet infos
        status = self.getAccountIsActive(address)

        if lock and not status:
            logger.info("The wallet %s is already locked", address)
            return None
        if not lock and status:
            logger.info("The wallet %s is already unlocked", address)
            return None

        # Get wallet infos
        acc_type = self.getAccountType(address)
        lim_m = self.getAccountCmLimitMin(address)
        lim_p = self.getAccountCmLimitMax(address)

        status = 1
        if lock:
            status = 0

        # prepare the data
        data = encodeAddressForTransaction(address)
        data += (
            encodeNumber(status)
            + encodeNumber(acc_type)
            + encodeNumber(int(lim_p * 100))
            + encodeNumber(int(lim_m * 100))
        )

        # send the transaction
        logger.info(
            "%s the wallet %s on server %s (%s)",
            "Locking" if lock else "Unlocking",
            address,
            self._currency_name,
            self.contracts[0],
        )
        return self._pyc3l.send_transaction(
            self.comchain._get_contract_fn_hexs("setAccountParam")[0],
            data, account)

    def pledge(self, account, address, amount, **kwargs):
        """Pledge a given amount to a Wallet on the current Currency (server)

        Parameters:
        account (eth_account import::Account): An account with admin permission on the current server. Will sign the transaction
        address (string): The public address of the wallet to be pledged (0x12345... format)
        amount (double): amount (in the current Currency) to be pledged to the wallet

        """
        # Check the admin
        self.checkAdmin(account.address)

        # Get wallet infos
        if not self.getAccountIsActive(address):
            logger.warn(
                "The target wallet %s is locked on server %s (%s)",
                address,
                self._currency_name,
                self.contracts[0],
            )

        # encode message if any
        if "message_to" in kwargs and kwargs["message_to"] != "":
            ciphered_message_to, public_key = self.encryptTransactionMessage(
                kwargs["message_to"], address=address
            )
        else:
            ciphered_message_to = ""

        # Prepare data
        amount_cent = int(100 * amount)
        data = encodeAddressForTransaction(address)
        data += encodeNumber(amount_cent)

        # send transaction
        logger.info(
            "Pledging %s to target wallet %s on server %s (%s) through end-point %s",
            amount,
            address,
            self._currency_name,
            self.contracts[0],
            self.endpoint,
        )
        return self._pyc3l.send_transaction(
            self.comchain._get_contract_fn_hexs("pledge")[0],
            data, account, "", ciphered_message_to
        )

    def delegate(self, account, address, amount):
        """Delegate a given amount of own money to an address

        Parameters:
        account: The account issuing the delegation
        address (string): The public address of the wallet to receive delegeation (0x12345... format)
        amount (double): amount (in the current Currency) to be delegated

        """

        # Prepare data
        amount_cent = int(100 * amount)
        data = encodeAddressForTransaction(address)
        data += encodeNumber(amount_cent)

        # send transaction
        logger.info(
            "Delegating %s to target wallet %s on server %s (%s) through end-point %s",
            amount,
            address,
            self._currency_name,
            self.contracts[0],
            self.endpoint,
        )
        return self._pyc3l.send_transaction(
            self.comchain._get_contract_fn_hexs("delegate")[0],
            data, account)

    def transferOnBehalfOf(self, account, address_from, address_to, amount, **kwargs):
        """Transfer amount money from address_from to address_to

        Note, this requires a delegation to be set up previously

        Parameters:
        account: The account issuing the order
        address_from (string): The address of the wallet to take money (0x12345... format)
        address_to (string): The address of the wallet to send money (0x12345... format)
        amount (double): amount (in the current Currency) to be transfered

        """

        # Prepare data
        data = encodeAddressForTransaction(address_from)
        data += encodeAddressForTransaction(address_to)
        data += encodeNumber(int(100 * amount))

        # prepare messages
        if "message_from" in kwargs and kwargs["message_from"] != "":
            ciphered_message_from, public_key = self.encryptTransactionMessage(
                kwargs["message_from"], address=address_from
            )
        else:
            ciphered_message_from = ""

        if "message_to" in kwargs and kwargs["message_to"] != "":
            ciphered_message_to, public_key = self.encryptTransactionMessage(
                kwargs["message_to"], address=address_to
            )
        else:
            ciphered_message_to = ""

        # send transaction
        logger.info(
            "Ask Transfer from %s to %s of %s",
            address_from,
            address_to,
            amount,
        )
        return self._pyc3l.send_transaction(
            self.comchain._get_contract_fn_hexs("transferNantOnBehalf")[0],
            data,
            account,
            ciphered_message_from,
            ciphered_message_to,
        )

    def __getattr__(self, label):
        return getattr(self.comchain, label)

